from torch.nn import Sequential
from torch import nn
from nalu.core.nalu_cell import NaluCell


class NaluLayer(nn.Module):
    def __init__(self, input_shape, output_shape, n_layers, hidden_shape):
        super().__init__()
        self.input_shape = input_shape
        self.output_shape = output_shape
        self.n_layers = n_layers
        self.hidden_shape = hidden_shape
        layers = [NaluCell(hidden_shape if n > 0 else input_shape,
                           hidden_shape if n < n_layers - 1 else output_shape) for n in range(n_layers)]
        self.model = Sequential(*layers)

    def forward(self, data):
        return self.model(data)



from torch import Tensor, exp, log, nn
from torch.nn.parameter import Parameter
from torch.nn.init import xavier_uniform_
from torch.nn.functional import sigmoid, linear
from .nac_cell import NacCell


class NaluCell(nn.Module):
    """Basic NALU unit implementation 
    from https://arxiv.org/pdf/1808.00508.pdf
    """

    def __init__(self, in_shape, out_shape):
        """
        in_shape: input sample dimension
        out_shape: output sample dimension
        """
        super().__init__()
        self.in_shape = in_shape
        self.out_shape = out_shape
        self.G = Parameter(Tensor(out_shape, in_shape))
        self.nac = NacCell(out_shape, in_shape)
        xavier_uniform_(self.G)
        self.eps = 1e-5
        self.register_parameter('bias', None)

    def forward(self, input):
        a = self.nac(input)
        g = sigmoid(linear(input, self.G, self.bias))
        ag = g * a
        log_in = log(abs(input) + self.eps)
        m = exp(self.nac(log_in))
        md = (1 - g) * m
        return ag + md



from torch import Tensor, nn
from torch.nn.parameter import Parameter
from torch.nn.init import xavier_uniform_
from torch.nn.functional import tanh, sigmoid, linear


class NacCell(nn.Module):
    """Basic NAC unit implementation 
    from https://arxiv.org/pdf/1808.00508.pdf
    """

    def __init__(self, in_shape, out_shape):
        """
        in_shape: input sample dimension
        out_shape: output sample dimension
        """
        super().__init__()
        self.in_shape = in_shape
        self.out_shape = out_shape
        self.W_ = Parameter(Tensor(out_shape, in_shape))
        self.M_ = Parameter(Tensor(out_shape, in_shape))
        xavier_uniform_(self.W_), xavier_uniform_(self.M_)
        self.register_parameter('bias', None)

    def forward(self, input):        
        W = tanh(self.W_) * sigmoid(self.M_)
        return linear(input, W, self.bias)



from io import open

from setuptools import find_packages, setup

with open('nalu/__init__.py', 'r') as f:
    for line in f:
        if line.startswith('__version__'):
            version = line.strip().split('=')[1].strip(' \'"')
            break
    else:
        version = '0.0.1'

with open('README.md', 'r', encoding='utf-8') as f:
    readme = f.read()

REQUIRES = ['torch', 'numpy']

setup(
    name='NALU',
    version=version,
    description='basic implementation of Neural arithmetic and logic units as described in arxiv.org/pdf/1808.00508.pdf',
    long_description=readme,
    author='Bharath G.S',
    author_email='royalkingpin@gmail.com',
    maintainer='Bharath G.S',
    maintainer_email='royalkingpin@gmail.com',
    url='https://github.com/bharathgs/NALU',
    license='MIT',

    keywords=[
        'NALU', 'ALU', 'neural', 'neural-networks', 'pytorch', 'NAC', 'torch', 'machine-learning'
    ],

    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: CPython',
    ],

    install_requires=REQUIRES,
    tests_require=['coverage', 'pytest'],

    packages=find_packages(),
)
